/*
Copyright (c) 2007 Health Market Science, Inc.

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
USA

You can contact Health Market Science at info@healthmarketscience.com
or at the following address:

Health Market Science
2700 Horizon Drive
Suite 200
King of Prussia, PA 19406
*/

package com.healthmarketscience.common.util;

import java.io.IOException;

/**
 * Extended Appendable interface which adds the ability for objects to append
 * themselves directly to the given Appendable instead of creating
 * intermediate CharSequence objects (Strings, etc).  This can make appends of
 * deep object hierarchies much more efficient.  An object can take advantage
 * of this facility by implementing the Appendee interface.
 * <p>
 * Additionally, this interface adds two convenience methods.
 * <ul>
 * <li>{@link #append(Object)} - a method to append Objects by calling
 *     String.valueOf(obj) on the object and appending to the result to the
 *     AppendableExt</li>
 * <li>{@link #append(Iterable,Object)} - a method to append an iteration
 *     (collection, etc) of objects separated by a delimiter</li>
 * </ul>
 * This class acts like the 
 * {@link java.util.Formatter}/{@link java.util.Formattable} facility without 
 * the extra overhead of parsing the format strings.  If complicated formatting
 * is needed for the generated strings, Formatter should be used instead.
 * </p>
 * <p>
 * Examples:
 * </p>
 * <pre>
 *
 * //
 * // *Without* using this interface.
 * //
 * public class Foo {
 *   private Bar b;
 *   public String toString() {
 *     return "Foo " + b;
 *   }
 * }
 * public class Bar {
 *   public String toString() {
 *     return "Bar";
 *   }
 * }
 *
 * Foo f = new Foo();
 * StringBuilder sb = new StringBuilder();
 *
 * // this will involve copying multiple strings before actual append!!!
 * sb.append(f);
 *
 * 
 * //
 * // *With* using this interface.
 * //
 * public class Foo implements Appendee {
 *   private Bar b;
 *   public void appendTo(AppendableExt app) throws IOException {
 *     app.append("Foo").append(b);
 *   }
 * }
 * public class Bar implements Appendee {
 *   public void appendTo(AppendableExt app) throws IOException {
 *     app.append("Bar");
 *   }
 * }
 *
 * Foo f = new Foo();
 * // note, Wrapper uses StringBuilder by default under the hood
 * AppendableExt.Wrapper sb = new AppendableExt.Wrapper();
 *
 * // this will involve no extra copies, both strings ("Foo", "Bar") will be
 * // written directly to the Appender
 * sb.append(f);
 *
 * </pre>
 *
 * @author James Ahlborn
 */
public interface AppendableExt extends Appendable {

  public AppendableExt append(char c)
    throws IOException;

  public AppendableExt append(CharSequence csq)
    throws IOException;
  
  public AppendableExt append(CharSequence csq, int start, int end)
    throws IOException;
  
  /**
   * Will call the appendTo() method on the given object.
   *
   * @param a object to append to this AppendableExt
   * @return this AppendableExt object
   * @throws IOException if the append fails
   */
  public AppendableExt append(Appendee a)
    throws IOException;

  /**
   * Will call append(String.valueOf(o)) with the given object (unless the
   * object is an Appendee or CharSequence in which case it will be passed to
   * the appropriate method).
   *
   * @param o object to append to this AppendableExt
   * @return this AppendableExt object
   * @throws IOException if the append fails
   */
  public AppendableExt append(Object o)
    throws IOException;

  /**
   * Will iterate the given Iterable and append each object separated by the
   * given delimiter.
   *
   * @param iable an Iterable object
   * @param delimiter delimiter to append between each object in the iable
   * @return this AppendableExt object
   * @throws IOException if the append fails
   */
  public AppendableExt append(Iterable<?> iable, Object delimiter)
    throws IOException;


  /**
   * @return the current "context" as set through {@link #setContext}, if any.
   */
  public Object getContext();

  /**
   * Sets the "context" that will be returned from subsequent
   * {@link #getContext} calls.  Useful for Appendee implementations that
   * want to change the behavior of nested, context-aware Appendee objects.
   *
   * @param newContext the new context for any subsequent append calls
   */
  public void setContext(Object newContext);
  
}
