/*
Copyright (c) 2007 Health Market Science, Inc.

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
USA

You can contact Health Market Science at info@healthmarketscience.com
or at the following address:

Health Market Science
2700 Horizon Drive
Suite 200
King of Prussia, PA 19406
*/

package com.healthmarketscience.common.util;

import java.util.Iterator;
import java.io.IOException;

/**
 * Simple wrapper for any Appendable that implements the extended
 * capability of AppendableExt.  Default constructors will use a
 * StringBuilder for the underlying Appendable.
 * <p>
 * Any IOExceptions thrown by the underlying Appendable are trapped, and
 * at any time the last IOException can be obtained using 
 * {@link #getIOException}.  This is only useful if the underlying Appendable
 * actually throws IOException, which StringBuilder does not do.
 * </p>
 * <p>
 * This class is not thread-safe.
 * </p>
 */
public class AppendableExtWrapper implements AppendableExt
{
  /** 
   * The actual Appendable getting the chars 
   */
  private final Appendable _app;

  /** 
   * The last IOException thrown by an append call
   */
  private IOException _ioException;

  /** 
   * Custom context for the appendable 
   */
  private Object _context;
    
  /** 
   * Initialize a new AppendableExtWrapper based on a StringBuilder.
   */
  public AppendableExtWrapper() {
    _app = new StringBuilder();
  }

  /** 
   * Initialize a new AppendableExtWrapper based on a StringBuilder, with a specified 
   * initial capacity.
   * 
   * @param capacity the initial capacity for the StringBuilder.
   */
  public AppendableExtWrapper(int capacity) {
    this(capacity, null);
  }

  public AppendableExtWrapper(int size, Object context) {
    _app = new StringBuilder(size);
    _context = context;
  }

  public AppendableExtWrapper(Appendable app) {
    _app = app;
  }

  public AppendableExtWrapper append(char c)
  {
    try {
      _app.append(c);
    } catch(IOException e) {
      _ioException = e;
    }
    return this;
  }
  
  public AppendableExtWrapper append(CharSequence s)
  {
    if(s instanceof Appendee) {
      return this.append((Appendee)s);
    }

    try {
      _app.append(s);
    } catch(IOException e) {
      _ioException = e;
    }
    return this;
  }
  
  /** 
   * Appends a range of characters from the given sequence to the 
   * underlying Appendable.  This may behave differently from 
   * {@link #append(CharSequence)}, since that method uses the appendTo 
   * method of its argument if it's an instance of Appendee.
   *
   * @param s the character sequence to append
   * @param start the index of the first character to append
   * @param end the index following the last character to append
   * @return this AppendableExtWrapper object
   */
  public AppendableExtWrapper append(CharSequence s, int start, int end)
  {
    try {
      _app.append(s, start, end);
    } catch(IOException e) {
      _ioException = e;
    }
    return this;
  }

  public AppendableExtWrapper append(Appendee a)
  {
    try {
      a.appendTo(this);
    } catch(IOException e) {
      _ioException = e;
    }
    return this;
  }

  public AppendableExtWrapper append(Object o)
  {
    if(o instanceof Appendee) {
      return this.append((Appendee)o);
    }
    if(o instanceof CharSequence) {
      return this.append((CharSequence)o);
    }

    try {
      _app.append(String.valueOf(o));
    } catch(IOException e) {
      _ioException = e;
    }
    return this;
  }

  public AppendableExtWrapper append(Iterable<?> iable, Object delimiter)
  {
    for(Iterator<?> iter = iable.iterator(); iter.hasNext(); ) {
      append(iter.next());
      if(iter.hasNext()) {
        append(delimiter);
      }
    }
    return this;
  }

  /**
   * Get the underlying Appendable for this wrapper. 
   *
   * @return the underlying Appendable for this wrapper.
   */
  public Appendable getAppendable() { return _app; }

  /**
   * Get the last IOException thrown by an append call, or {@code null} if 
   * no IOException was caught.
   *
   * @return the last IOException thrown by an append call, or {@code null}
   *         if no IOException was caught.
   */
  public IOException getIOException() { return _ioException; }

  /**
   * Get the result of calling toString() on the underlying Appendable.
   *
   * @return the result of calling toString() on the underlying Appendable.
   */
  @Override
  public String toString() {
    return _app.toString();
  }

  public Object getContext() {
    return _context;
  }

  public void setContext(Object newContext) {
    _context = newContext;
  }
  
}
