/*
Copyright (c) 2007 Health Market Science, Inc.

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
USA

You can contact Health Market Science at info@healthmarketscience.com
or at the following address:

Health Market Science
2700 Horizon Drive
Suite 200
King of Prussia, PA 19406
*/

package com.healthmarketscience.common.util.resource;

import com.healthmarketscience.common.util.URLProtocolHandlerUtil;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLStreamHandler;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Classpath resource URLHandler provides easy access to resources via the
 * classloader system. URL's take the form of <PRE>
 *  resource:/com/company/path/to/file.txt
 * </PRE>
 *
 * @author Tim McCune
 * @see java.net.URL
 */
public class Handler extends URLStreamHandler {

  private static final Log LOG = LogFactory.getLog(Handler.class);
  private static final String PROTOCOL = "resource";
  
  static {
    try {
      URLProtocolHandlerUtil.registerProtocolHandler(PROTOCOL, new Handler());
    } catch (Error e) {
      LOG.warn("Can't register protocol handler", e);
      throw e;
    } catch (RuntimeException e) {
      LOG.warn("Can't register protocol handler", e);
      throw e;
    }
  }

  /** 
   * Classes that need to handle resource-style URLs can call this message to
   * ensure that this class is initialized.
   */
  public static void init() {
    // Don't do anything.  This method is just called to initialize class.
  }
  
  /**
   * Parse a string into a URL object. Host and port are not changed. Only
   * the file attribute is changed.
   */
  @Override
  protected void parseURL(URL u,
  String spec,
  int start,
  int limit) {

    // Args to setURL are:
    //   URL u,
    //   String protocol,
    //   String host,
    //   int port,
    //   String authority,
    //   String userInfo,
    //   String file,
    //   String query,
    //   String ref
    super.setURL(u, PROTOCOL, null, -1, null, null, 
    spec.substring(start, limit), null, null);
    
  }
  
  
  /**
   * Opens a connection to the object referenced by the URL argument.
   */
  @Override
  protected URLConnection openConnection(URL u)
  throws IOException {
    
    URL res = convertToSystemURL(u);
    
    if (res == null) {
      throw new FileNotFoundException(u + " not found");
    }
    return res.openConnection();
  }
  
  
  public URL convertToSystemURL(URL u) {
    String path = u.getFile();
    while (path.charAt(0) == '/') {
      path = path.substring(1);
      // skip preceding slashes
    }
    
    ClassLoader loader = Thread.currentThread().getContextClassLoader();
    URL res = ((loader != null) ? loader.getResource(path) : null);
    
    if (res == null) {
      res = u.getClass().getResource( path );
    }
    if (res == null) {
      res = ClassLoader.getSystemResource(path);
    }
    while (res == null && loader != null) {
      loader = loader.getParent();
      if (loader != null) {
        res = loader.getResource(path);
      }
    }
    
    return res;
  }
  
}
